/*
 * This file specifically contains a way to fetch traps from
 * the snmptrapd mysql interface. It would be possible to
 * implement a get_traps that fetches its traps from somewhere else,
 * and use that instead of this.
 */

#include <string.h>

#include "dbbackend.h"
#include "lua.h"
#include "log.h"

static proc_field *
parse_field_value(char *varval)
{
	proc_field *field = calloc(1, sizeof(proc_field));
	if (!strncmp(varval, "STRING: ", strlen("STRING: "))
			|| !strncmp(varval, "Hex-STRING: ", strlen("Hex-STRING: ")))
	{
		field->type = FIELD_STRING;
		char *start =  strchr(varval, ':');
		start += 3; // past ": \""
		field->value.string = strndup(start, strlen(start)-1);
	}
	else if (!strncmp(varval, "IpAddress: ", strlen("IpAddress: "))
			|| !strncmp(varval, "OID: ", strlen("OID: ")))
	{
		field->type = FIELD_STRING;
		varval =  strchr(varval, ':');
		varval += 2; // past ": "
		field->value.string = strdup(varval);
	}
	else if (!strcmp(varval, "\"\""))
	{
		field->type = FIELD_STRING;
		field->value.string = strdup("");
	}
	else if (!strncmp(varval, "Timeticks: ", strlen("Timeticks: ")))
	{
		field->type = FIELD_TIME;
		field->value.time = atoi(strchr(varval, '(') + 1);
	}
	else if (!strncmp(varval, "INTEGER: ", strlen("INTEGER"))
			|| !strncmp(varval, "Gauge32: ", strlen("Gauge32: "))
			|| !strncmp(varval, "Counter32: ", strlen("Counter32: "))
			|| !strncmp(varval, "Counter64: ", strlen("Counter64: ")))
	{
		field->type = FIELD_NUMERIC;
		field->value.number = atoi(strchr(varval, ':') + 2);
	}
	else if (!strncmp(varval, "BOOLEAN: ", strlen("BOOLEAN: ")))
	{
		field->type = FIELD_BOOLEAN;
		field->value.boolean = !strncmp(varval, "BOOLEAN: TRUE", strlen("BOOLEAN: TRUE"));
	}
	else if (!strncmp(varval, "NULL", strlen("NULL"))) {
		field->type = FIELD_NULL;
		field->value.number = 0;
	}
	else {
		log_msg(LOG_ERR, "Don't understand value '%s' - ignoring", varval == NULL ? "(NULL)" : varval);
		return NULL;
	}
	return field;
}


static int
set_fields(dbi_conn conn, proc_trap *trap)
{
	dbi_result result = dbi_conn_queryf(conn,
			"SELECT oid, value "
			"FROM varbinds "
			"WHERE trap_id = %d",
			trap->id);
	if (!result)
		return 1;
	while (dbi_result_next_row(result)) {
		char *oid = dbi_result_get_string_copy(result, "oid");
		const char *tmp = (const char *)dbi_result_get_binary(result, "value");
		char *varval = strndup(tmp, dbi_result_get_field_length(result, "value") + 1);
		if (!strcmp(varval, "ERROR")) {
			const char *msg;
			dbi_conn_error(conn, &msg);
			log_msg(LOG_ERR, "Error fetching varbind value: %s", msg);
		}
		proc_field *field = parse_field_value(varval);
		free(varval);
		if (field == NULL)
			continue;
		field->oid = oid;

		if (!strcmp(field->oid, ".0")) {
			// We don't understand that we use transactions for the sql backend
			// so we hack this pseudo-oid in whenever we've received a complete
			// trap.
			free(field);
			field = NULL;
		} else if (!strcmp(field->oid, ".1.3.6.1.2.1.1.3.0")) {
			// where trap->timestamp is set from,
			// but it already is implicitly
			free(field->oid);
			free(field);
			field = NULL;
		} else if (!strcmp(field->oid, ".1.3.6.1.6.3.1.1.4.1.0")) {
			// where trap->oid is set from,
			// but it already is implicitly
			free(field->oid);
			free(field->value.string);
			free(field);
			field = NULL;
		}
		else {
			proc_field *last = trap->fields;
			if (!last) {
				trap->fields = field;
			}
			else {
				while (last->next) {
					last = last->next;
				}
				last->next = field;
			}
		}
	}
	dbi_result_free(result);
	return 0;
}

proc_trap *
get_traps(dbi_conn conn, int *errcode, const char **errmsg_dest)
{
	dbi_result result;
	proc_trap *prev = NULL, *trap = NULL;
	const char *query_trap = "SELECT trap_id, unix_timestamp(date_time) as timestamp, host, snmpTrapOID \
			 	  FROM notifications \
			 	  ORDER BY trap_id DESC LIMIT 100";

	result = dbi_conn_query(conn, query_trap);
	if (!result) {
		*errcode = dbi_conn_error(conn, errmsg_dest);
		return NULL;
	}
	while (dbi_result_next_row(result)) {
		prev = trap;
		trap = calloc(1, sizeof(proc_trap));
		trap->id = dbi_result_get_uint(result, "trap_id");
		/* investigation shows that calling "long" here just retrieves 0 */
		trap->timestamp = dbi_result_get_longlong(result, "timestamp");
		trap->host = dbi_result_get_string_copy(result, "host");
		trap->oid = dbi_result_get_string_copy(result, "snmpTrapOID");

		if (prev)
			trap->next = prev;
		set_fields(conn, trap);
	}
	dbi_result_free(result);

	return trap;
}

int
delete_trap(dbi_conn conn, proc_trap *trap)
{
	dbi_result dbr = dbi_conn_queryf(conn,
			"DELETE FROM notifications WHERE trap_id = %d",
			trap->id);
	dbi_result_free(dbr);
	return 0;
}
